// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file DataCollector.h */

#ifndef AMDIS_DATACOLLECTOR_H
#define AMDIS_DATACOLLECTOR_H

#include <list>
#include <vector>
#include "ElementInfo.h"
#include "VertexInfo.h"
#include "PeriodicInfo.h"
#include "Mesh.h"
#include "AMDiS_fwd.h"

namespace AMDiS {

  /** 
   * \ingroup Output
   *
   * \brief
   * Provides data collecting of element, vertex and value data for file writer.
   */
  class DataCollector 
  {
  public:
    /// Constructor
    DataCollector(const FiniteElemSpace *feSpace,
		  DOFVector<double> *values,
		  int level = -1,
		  Flag traverseFlag = Mesh::CALL_LEAF_EL,
		  bool (*writeElem)(ElInfo*) = NULL);

    ~DataCollector();

    /// Fills the DataCollector with all possible datas.
    void fillAllData();

    /// Returns list with element information.
    std::list<ElementInfo>* getElementInfos();

    /// Returns vector with vertex information.
    DOFVector< std::list<VertexInfo> >* getVertexInfos();

    /// Returns the finite element space of the problem.
    const FiniteElemSpace* getFeSpace();

    /// Returns vector with value information.
    DOFVector<double>* getValues();

    /// Returns vector with information about dof coordinates.
    DOFVector< std::list<WorldVector<double> > >* getDofCoords();

    /// Returns vector with information about interpolation point indexing.
    DOFVector<int>* getInterpPointInd();

    ///
    DOFVector< std::list<WorldVector<double> > >* getInterpPointCoords();

    /// Returns list of interpolation point information.
    std::vector< std::vector<int> >* getInterpPoints();

    /// Returns list of information about periodics.
    std::list<PeriodicInfo>* getPeriodicInfos();

    /// Returns the number of vertices.
    int getNumberVertices();

    /// Returns the number of elements.
    int getNumberElements();

    /// Returns the number of interpolation points.
    int getNumberInterpPoints();

    // Returns the number of connections.
    int getNumberConnections();
      
    /// Returns the mesh of the problem.
    Mesh* getMesh();

    void setMesh(Mesh *mesh) {
      mesh_ = mesh;
    }

  protected:  
    /// Start collecting element and vertex data of the problem.
    void startCollectingElementData();

    /// Start collecting value data of the problem.
    void startCollectingValueData();

    /// Start collecting periodic data of the problem.
    void startCollectingPeriodicData();

    /// Adds information about one element and its vertices.
    void addElementData(ElInfo* elInfo);

    /// Adds value information of one element.
    void addValueData(ElInfo *elInfo);

    /// Adds information about interpolation points of vertices.
    void addInterpData(ElInfo *elInfo);

    /// Adds value information of one element.
    void addPeriodicData(ElInfo *elInfo);

    /// Vector with vertex values
    DOFVector<double> *values_;

    /// Level information for traversing the mesh.
    int level_;

    /// Flags for traversing the mesh.
    Flag traverseFlag_;

    ///
    const FiniteElemSpace *feSpace_;

    /// Mesh that should be written
    Mesh *mesh_;
      
    /// DOFAdmin of values
    DOFAdmin *localAdmin_;

    /// vertex pre-dofs
    int nPreDofs_;

    /// Number of vertices.
    int nVertices_;

    /// Number of elements.
    int nElements_;

    /// Total number of interpolation points.
    int nInterpPoints_;

    /// Number of connections in periodic problems.
    int nConnections_;

    /// Dimension of \ref mesh
    int dim_;
           
    /// Maps internal element indices to global output indices.
    std::map<int, int> outputIndices_;

    /// Global interpolation point indexing
    DOFVector<int> *interpPointInd_;

    /// Stores for each simplex the interpolation points.
    std::vector< std::vector<int> > interpPoints_;

    /** \brief
     * Stores for each DOF a list of its coordinates. If there are now periodic
     * boundaries than there is also only one coordinate per DOF.
     */
    DOFVector< std::list<WorldVector<double> > > *interpPointCoords_;

    /// list of coords for each dof
    DOFVector< std::list<WorldVector<double> > > *dofCoords_;

    /// List that stores an ElementInfo for each element.
    std::list<ElementInfo> elements_;

    /// List stat stores information about all periodics.
    std::list<PeriodicInfo> periodicInfos_;

    /// Stores a list of vertex infos for each dof.
    DOFVector< std::list<VertexInfo> > *vertexInfos_;

    /** \brief
     * periodicConnections[i][j] stores whether the connection at side j of 
     * the element with output index i has already been written.
     */
    std::vector<DimVec<bool> > periodicConnections_;

    /// Stores if element data was collected before.
    bool elementDataCollected_;

    /// Stores if value data was collected before.   
    bool valueDataCollected_;

    /// Stores if periodic data was collected before.
    bool periodicDataCollected_;

    /// Pointer to a function which decides whether an element is considered.
    bool (*writeElem_)(ElInfo*);

    /// Temporary variable used in functions addValueData() and addInterpData().
    DegreeOfFreedom *localDOFs_;
      
    /// Temporary variable used in functions addValueData() and addInterpData().
    BasisFunction *basisFcts_;

    /// Temporary variable used in functions addValueData() and addInterpData().
    int nBasisFcts_;

    /// Temporary variable used in function \ref addValueData.
    WorldVector<double> *vertexCoords;
  };
}

#endif
